#include "testbase.h"
#include "vector.h"

#include <algorithm>
#include <cstdlib>
#include <iostream>
#include <vector>


using namespace std;


// Change undef to define to use some STL algorithms on your Vector<T>
#undef TEST_STL_ALGORITHMS

// Change undef to define if you implement iterators for Vector<bool>
#undef TEST_VECBOOL_ITERS


/*===========================================================================
 * HELPER FUNCTIONS
 *
 * These are used by the test suite.
 */

string make_string(int num_chars) {
    assert(num_chars > 0);

    string s;
    
    for (int i = 0; i < num_chars; i++)
        s += (char) ('A' + rand() % 26);
    
    return s;
}

string make_string(int min_chars, int max_chars) {
    assert(min_chars > 0);
    assert(max_chars >= min_chars);
    return make_string(min_chars + (1 + max_chars - min_chars) % rand());
}


/*===========================================================================
 * TEST FUNCTIONS
 *
 * These are called by the main() function at the end of this file.
 */


Vector<int> makeTestVector(int n) {
    Vector<int> v;
    for (int i = 0; i < n; i++)
        v.push_back(i);
    
    return v;
}


/*! Test the Vector constructors. */
void test_constructors(TestContext &ctx) {
    ctx.DESC("Vector size constructor");

    Vector<int> v1;
    ctx.CHECK(v1.size() == 0);
    ctx.CHECK(v1.capacity() == 0);
    
    Vector<float> v2(10);
    ctx.CHECK(v2.size() == 10);
    
    ctx.CHECK(v2.capacity() >= 10);

    ctx.result();

    ctx.DESC("Vector copy constructor / index operator");

    Vector<int> vo;
    vo.push_back(3);
    vo.push_back(5);
    vo.push_back(4);

    Vector<int> v3(vo);

    ctx.CHECK(vo[0] == 3);
    ctx.CHECK(vo[1] == 5);
    ctx.CHECK(vo[2] == 4);

    ctx.CHECK(v3[0] == 3);
    ctx.CHECK(v3[1] == 5);
    ctx.CHECK(v3[2] == 4);

    v3[1] = -2;

    ctx.CHECK(vo[0] == 3);
    ctx.CHECK(vo[1] == 5);
    ctx.CHECK(vo[2] == 4);

    ctx.CHECK(v3[0] == 3);
    ctx.CHECK(v3[1] == -2);
    ctx.CHECK(v3[2] == 4);
    
    ctx.result();

    ctx.DESC("Vector move constructor");
    
    Vector<int> v4 = makeTestVector(30);
    ctx.CHECK(v4.size() == 30);
    for (int i = 0; i < 30; i++)
        ctx.CHECK(v4[i] == i);
    
    ctx.result();

    ctx.DESC("Vector list-initialization constructor");

    Vector<int> v5 = {10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0};

    ctx.CHECK(v5.size() == 11);
    for (int i = 0; i < v5.size(); i++)
        ctx.CHECK(v5[i] == v5.size() - i - 1);

    Vector<string> v6 = {"red", "orange", "yellow", "green", "blue", "violet"};
    ctx.CHECK(v6.size() == 6);
    ctx.CHECK(v6[0] == "red");
    ctx.CHECK(v6[1] == "orange");
    ctx.CHECK(v6[2] == "yellow");
    ctx.CHECK(v6[3] == "green");
    ctx.CHECK(v6[4] == "blue");
    ctx.CHECK(v6[5] == "violet");

    ctx.result();
}


void test_bad_indexes(TestContext &ctx) {
    Vector<int> v;

    ctx.DESC("Vector indexing reports bad indexes via out_of_range");

    try {
        v[0] = 5;
        ctx.CHECK(false);  // Shouldn't get here.
    } catch (out_of_range &e) {
        ctx.CHECK(true);   // PASS
    } catch (...) {
        ctx.CHECK(false);  // Caught unexpected exception!
    }

    try {
        v[-1];
        ctx.CHECK(false);  // Shouldn't get here.
    } catch (out_of_range &e) {
        ctx.CHECK(true);   // PASS
    } catch (...) {
        ctx.CHECK(false);  // Caught unexpected exception!
    }

    // This should work now.
    v.push_back(5);
    ctx.CHECK(v[0] == 5);
 
    try {
        v[1] = 5;
        ctx.CHECK(false);  // Shouldn't get here.
    } catch (out_of_range &e) {
        ctx.CHECK(true);   // PASS
    } catch (...) {
        ctx.CHECK(false);  // Caught unexpected exception!
    }

    ctx.result();
}


void test_assignment(TestContext &ctx) {
    Vector<int> v1, v2;
    
    ctx.DESC("Vector assignment operator");
    
    v1 = makeTestVector(30);     // Move-assignment
    v2 = v1;                     // Copy-assignment
    ctx.CHECK(v2.size() == 30);
    for (int i = 0; i < 30; i++) {
        ctx.CHECK(v1[i] == i);
        ctx.CHECK(v2[i] == i);
    }
    v2[12] = -63;
    ctx.CHECK(v1[12] == 12 && v2[12] == -63);

    Vector<int> v3;
    for (int i = 0; i < 1000; i++)
        v3.push_back(i);

    v3 = v1;
    ctx.CHECK(v3.size() == 30);
    for (int i = 0; i < 30; i++) {
        ctx.CHECK(v3[i] == i);
        ctx.CHECK(v1[i] == i);
    }
    v3[12] = -63;
    ctx.CHECK(v1[12] == 12 && v3[12] == -63);
    ctx.result();
}


void test_push_back(TestContext &ctx) {
    ctx.DESC("push_back() stores values and grows Vector properly");
    
    Vector<int> v;
    int i;
    const int NUMVALS = 70000;
    
    // Put 20,000 ints into the vector.
    for (i = 0; i < NUMVALS; i++)
        v.push_back(i);
    
    // Verify that the values look correct through array indexing.
    for (i = 0; i < NUMVALS; i++)
        ctx.CHECK(v[i] == i);

    ctx.result();
}


void test_forward_iterators(TestContext &ctx) {
    Vector<int> v;
    const Vector<int> &cv(v);  // A const reference to v
    int i;

    for (i = 0; i < 10; i++)
        v.push_back(i);

    ctx.DESC("basic usage of forward iterators");

    // Check forward iterator, using pre-increment
    i = 0;
    for (Vector<int>::iterator it = v.begin(); it != v.end(); ++it, i++)
        ctx.CHECK(*it == i);

    // Check forward iterator, using post-increment
    i = 0;
    for (Vector<int>::iterator it = v.begin(); it != v.end(); it++, i++)
        ctx.CHECK(*it == i);

    ctx.result();

    ctx.DESC("basic usage of forward const-iterators");

    // Check forward iterator, using pre-increment
    i = 0;
    for (Vector<int>::const_iterator it = cv.begin(); it != cv.end(); ++it, i++)
        ctx.CHECK(*it == i);

    // Check forward iterator, using post-increment
    i = 0;
    for (Vector<int>::const_iterator it = cv.begin(); it != cv.end(); it++, i++)
        ctx.CHECK(*it == i);

    ctx.result();

    // Check the -> member-of-pointer operator

    ctx.DESC("pointer-to-member operator on forward iterator types");

    Vector<string> vs;
    const Vector<string> &cvs(vs);  // A const reference to vs

    for (i = 0; i < 10; i++)
        vs.push_back(make_string(10, 40));

    for (Vector<string>::iterator it = vs.begin(); it != vs.end(); it++)
        ctx.CHECK(it->size() == (*it).size());

    for (Vector<string>::const_iterator it = cvs.begin(); it != cvs.end(); it++)
        ctx.CHECK(it->size() == (*it).size());

    ctx.result();
}


void test_reverse_iterators(TestContext &ctx) {
    Vector<int> v;
    const Vector<int> &cv(v);  // A const reference to v
    int i;

    for (i = 0; i < 10; i++)
        v.push_back(i);

    ctx.DESC("basic usage of reverse iterators");

    // Check reverse iterator, using pre-increment
    i = 9;
    for (Vector<int>::reverse_iterator it = v.rbegin(); it != v.rend(); ++it, i--)
        ctx.CHECK(*it == i);

    // Check reverse iterator, using post-increment
    i = 9;
    for (Vector<int>::reverse_iterator it = v.rbegin(); it != v.rend(); it++, i--)
        ctx.CHECK(*it == i);

    ctx.result();

    ctx.DESC("basic usage of reverse const-iterators");

    // Check reverse iterator, using pre-increment
    i = 9;
    for (Vector<int>::const_reverse_iterator it = cv.rbegin(); it != cv.rend(); ++it, i--)
        ctx.CHECK(*it == i);

    // Check reverse iterator, using post-increment
    i = 9;
    for (Vector<int>::const_reverse_iterator it = cv.rbegin(); it != cv.rend(); it++, i--)
        ctx.CHECK(*it == i);

    ctx.result();

    // Check the -> member-of-pointer operator

    ctx.DESC("pointer-to-member operator on reverse iterator types");

    Vector<string> vs;
    const Vector<string> &cvs(vs);  // A const reference to vs

    for (i = 0; i < 10; i++)
        vs.push_back(make_string(10, 40));

    for (Vector<string>::reverse_iterator it = vs.rbegin(); it != vs.rend(); it++)
        ctx.CHECK(it->size() == (*it).size());

    for (Vector<string>::const_reverse_iterator it = cvs.rbegin(); it != cvs.rend(); it++)
        ctx.CHECK(it->size() == (*it).size());

    ctx.result();
}


#ifdef TEST_STL_ALGORITHMS


void test_reverse(TestContext &ctx) {
    ctx.DESC("std::reverse() works with Vector");
    
    Vector<int> v;
    int i;
    const int NUMVALS = 100;
    
    for (i = 0; i < NUMVALS; i++)
        v.push_back(i);
    
    std::reverse(v.begin(), v.end());
    
    for (i = 0; i < NUMVALS; i++)
        ctx.CHECK(v[i] == NUMVALS - i - 1);
    
    // Verify that the values look correct through the iterator.
    Vector<int>::iterator it = v.begin();
    i = 0;
    while (it != v.end()) {
        ctx.CHECK(*it == NUMVALS - i - 1);
        i++;
        it++;
    }
    ctx.CHECK(i == NUMVALS);
    
    std::reverse(v.begin(), v.end());
    
    for (i = 0; i < NUMVALS; i++)
        ctx.CHECK(v[i] == i);
    
    ctx.result();
}


void test_sort_ints(TestContext &ctx) {
    const int NUMVALS = 10000;
    int i, last_i;
    
    ctx.DESC("std::sort() works with Vector<int>");
    
    Vector<int> v;

    // Insert random numbers
    for (i = 0; i < NUMVALS; i++)
        v.push_back(rand() % NUMVALS);

    // Sort it!
    std::sort(v.begin(), v.end());

    // Check the order
    last_i = v[0];
    for (i = 1; i < NUMVALS; i++) {
        ctx.CHECK(last_i <= v[i]);
        last_i = v[i];
    }

    // Verify that the values look correct through the iterator.
    Vector<int>::iterator it = v.begin();
    
    last_i = *it;
    it++;
    
    i = 1;
    
    while (it != v.end()) {
        ctx.CHECK(last_i <= *it);

        last_i = *it;
        it++;
        
        i++;
    }
    ctx.CHECK(i == NUMVALS);
    
    ctx.result();
}


void test_sort_strings(TestContext &ctx) {
    const int NUMVALS = 10000;
    int i;
    string last_i;
    
    ctx.DESC("std::sort() works with Vector<string>");
    
    Vector<string> v;

    // Insert random numbers
    for (i = 0; i < NUMVALS; i++)
        v.push_back(make_string(4, 12));

    // Sort it!
    std::sort(v.begin(), v.end());

    // Check the order
    last_i = v[0];
    for (i = 1; i < NUMVALS; i++) {
        ctx.CHECK(last_i <= v[i]);
        last_i = v[i];
    }

    // Verify that the values look correct through the iterator.
    Vector<string>::iterator it = v.begin();
    
    last_i = *it;
    it++;
    
    i = 1;
    
    while (it != v.end()) {
        ctx.CHECK(last_i <= *it);

        last_i = *it;
        it++;
        
        i++;
    }
    ctx.CHECK(i == NUMVALS);
    
    ctx.result();
}

#endif // TEST_STL_ALGORITHMS


/*! Test Vectors of bool */
void test_vecbool(TestContext &ctx) {
    ctx.DESC("Vector<bool> - simple constructors / size() / capacity()");

    Vector<bool> v1;
    ctx.CHECK(v1.size() == 0);
    ctx.CHECK(v1.capacity() == 0);

    Vector<bool> v2(10);
    ctx.CHECK(v2.size() == 10);
    ctx.CHECK(v2.capacity() >= 10);

    ctx.result();

    ctx.DESC("Vector<bool> - list initializer / indexed reads / const reads");

    Vector<bool> vl = {true, false, false, true, false, true, true, false};
    const Vector<bool> &vlc(vl);
    ctx.CHECK(vl.size() == 8);
    ctx.CHECK(vl.capacity() >= 8);

    ctx.CHECK( vl[0]);
    ctx.CHECK(!vl[1]);
    ctx.CHECK(!vl[2]);
    ctx.CHECK( vl[3]);
    ctx.CHECK(!vl[4]);
    ctx.CHECK( vl[5]);
    ctx.CHECK( vl[6]);
    ctx.CHECK(!vl[7]);

    ctx.CHECK( vlc[0]);
    ctx.CHECK(!vlc[1]);
    ctx.CHECK(!vlc[2]);
    ctx.CHECK( vlc[3]);
    ctx.CHECK(!vlc[4]);
    ctx.CHECK( vlc[5]);
    ctx.CHECK( vlc[6]);
    ctx.CHECK(!vlc[7]);

    ctx.result();

    ctx.DESC("Vector<bool> - copy constructor / indexed writes / push_back()");

    Vector<bool> vo;
    vo.push_back(false);
    vo.push_back(true);
    vo.push_back(true);

    Vector<bool> vc(vo);

    ctx.CHECK(!vo[0]);
    ctx.CHECK( vo[1]);
    ctx.CHECK( vo[2]);

    ctx.CHECK(!vc[0]);
    ctx.CHECK( vc[1]);
    ctx.CHECK( vc[2]);

    vo[0] = true;
    vc[1] = false;

    ctx.CHECK( vo[0]);
    ctx.CHECK( vo[1]);
    ctx.CHECK( vo[2]);

    ctx.CHECK(!vc[0]);
    ctx.CHECK(!vc[1]);
    ctx.CHECK( vc[2]);

    ctx.result();

    ctx.DESC("Vector<bool> - larger bit-vector");

    // Make a larger vector of random bools - something that will take
    // multiple blocks, and probably require resizing too

    constexpr int NUM_BITS = 1000;
    Vector<bool> vb;
    std::vector<bool> svb;
    for (int i = 0; i < NUM_BITS; i++) {
        bool b = rand() % 2;

        vb.push_back(b);
        svb.push_back(b);

        ctx.CHECK(vb[i] == b);
        ctx.CHECK(svb[i] == b);
    }

    for (int i = 0; i < NUM_BITS; i++)
        ctx.CHECK(vb[i] == svb[i]);

    ctx.result();
}


#ifdef TEST_VECBOOL_ITERS

/*! Test Vectors of bool - iterators */
void test_vecbool_iterators(TestContext &ctx) {
    constexpr int NUM_BITS = 10000;

    ctx.DESC("Vector<bool> - iterators");

    // Make a random vector of bools.
    Vector<bool> v;
    for (int i = 0; i < NUM_BITS; i++)
        v.push_back(rand() % 2);

    int i = 0;
    for (Vector<bool>::iterator it = v.begin(); it != v.end(); it++, i++)
        ctx.CHECK(*it == v[i]);

    i = 0;
    for (Vector<bool>::iterator it = v.begin(); it != v.end(); ++it, i++)
        ctx.CHECK(*it == v[i]);

    i = 0;
    for (Vector<bool>::reverse_iterator rit = v.rbegin(); rit != v.rend();
         rit++, i++) {
        ctx.CHECK(*rit == v[NUM_BITS - i - 1]);
    }

    i = 0;
    for (Vector<bool>::reverse_iterator rit = v.rbegin(); rit != v.rend();
         ++rit, i++) {
        ctx.CHECK(*rit == v[NUM_BITS - i - 1]);
    }

    //TODO:  CONST ITERATORS?

    ctx.result();
}

#endif // TEST_VECBOOL_ITERS


/*! This program is a simple test-suite for the Vector class. */
int main() {

    srand(123456);  // So tests with rand() are repeatable
  
    cout << "Lab 3:  Testing complete Vector implementation and "
            "Vector<bool> specialization\n\n";

    TestContext ctx(cout);

    test_constructors(ctx);
    test_bad_indexes(ctx);
    test_assignment(ctx);
    test_push_back(ctx);
    test_forward_iterators(ctx);
    test_reverse_iterators(ctx);

#ifdef TEST_STL_ALGORITHMS
    test_reverse(ctx);
    test_sort_ints(ctx);
    test_sort_strings(ctx);
#endif

    test_vecbool(ctx);

#ifdef TEST_VECBOOL_ITERS
    test_vecbool_iterators(ctx);
#endif

    // Return 0 if everything passed, nonzero if something failed.
    return !ctx.ok();
}
